@@ -7,9 +7,9 @@ package tun
 
 import (
 	"errors"
+	"fmt"
 	"os"
 	"sync"
-	"syscall"
 	"time"
 	"unsafe"
 
@@ -48,7 +48,7 @@ type NativeTun struct {
 	wrBuff       *exchgBufWrite
 	events       chan TUNEvent
 	errors       chan error
-	forcedMtu    int
+	forcedMTU    int
 }
 
 func packetAlign(size uint32) uint32 {
@@ -60,32 +60,35 @@ func packetAlign(size uint32) uint32 {
 // adapter with the same name exist, it is reused.
 //
 func CreateTUN(ifname string) (TUNDevice, error) {
+	var err error
+	var wt *wintun.Wintun
+
 	// Does an interface with this name already exist?
-	wt, err := wintun.GetInterface(ifname, 0)
+	wt, err = wintun.GetInterface(ifname, 0)
 	if wt == nil {
-		// Interface does not exist or an error occured. Create one.
+		// Interface does not exist or an error occurred. Create one.
 		wt, _, err = wintun.CreateInterface("WireGuard Tunnel Adapter", 0)
 		if err != nil {
-			return nil, errors.New("Creating Wintun adapter failed: " + err.Error())
+			return nil, fmt.Errorf("wintun.CreateInterface: %v", err)
 		}
 	} else if err != nil {
 		// Foreign interface with the same name found.
 		// We could create a Wintun interface under a temporary name. But, should our
-		// proces die without deleting this interface first, the interface would remain
+		// process die without deleting this interface first, the interface would remain
 		// orphaned.
-		return nil, err
+		return nil, fmt.Errorf("wintun.GetInterface: %v", err)
 	}
 
 	err = wt.SetInterfaceName(ifname)
 	if err != nil {
 		wt.DeleteInterface(0)
-		return nil, err
+		return nil, fmt.Errorf("wintun.SetInterfaceName: %v", err)
 	}
 
 	err = wt.FlushInterface()
 	if err != nil {
 		wt.DeleteInterface(0)
-		return nil, errors.New("Flushing interface failed: " + err.Error())
+		return nil, fmt.Errorf("wintun.FlushInterface: %v", err)
 	}
 
 	return &NativeTun{
@@ -94,7 +97,7 @@ func CreateTUN(ifname string) (TUNDevice, error) {
 		wrBuff:    &exchgBufWrite{},
 		events:    make(chan TUNEvent, 10),
 		errors:    make(chan error, 1),
-		forcedMtu: 1500,
+		forcedMTU: 1500,
 	}, nil
 }
 
@@ -218,12 +221,12 @@ func (tun *NativeTun) Close() error {
 }
 
 func (tun *NativeTun) MTU() (int, error) {
-	return tun.forcedMtu, nil
+	return tun.forcedMTU, nil
 }
 
 //TODO: This is a temporary hack. We really need to be monitoring the interface in real time and adapting to MTU changes.
-func (tun *NativeTun) ForceMtu(mtu int) {
-	tun.forcedMtu = mtu
+func (tun *NativeTun) ForceMTU(mtu int) {
+	tun.forcedMTU = mtu
 }
 
 func (tun *NativeTun) Read(buff []byte, offset int) (int, error) {
@@ -273,7 +276,7 @@ func (tun *NativeTun) Read(buff []byte, offset int) (int, error) {
 					retries--
 					continue
 				}
-				if ok && pe.Err == syscall.Errno(6) /*windows.ERROR_INVALID_HANDLE*/ {
+				if ok && pe.Err == windows.ERROR_HANDLE_EOF {
 					tun.closeTUN()
 					break
 				}
@@ -301,7 +304,7 @@ func (tun *NativeTun) Flush() error {
 		}
 
 		// Flush write buffer.
-		retries := retryTimeout * retryRate
+		retries := 1000
 		for {
 			_, err = file.Write(tun.wrBuff.data[:tun.wrBuff.offset])
 			tun.wrBuff.packetNum = 0
@@ -312,11 +315,10 @@ func (tun *NativeTun) Flush() error {
 					return os.ErrClosed
 				}
 				if retries > 0 && ok && pe.Err == windows.ERROR_OPERATION_ABORTED {
-					time.Sleep(time.Second / retryRate)
 					retries--
 					continue
 				}
-				if ok && pe.Err == syscall.Errno(6) /*windows.ERROR_INVALID_HANDLE*/ {
+				if ok && pe.Err == windows.ERROR_HANDLE_EOF {
 					tun.closeTUN()
 					break
 				}
@@ -371,3 +373,10 @@ func (tun *NativeTun) Write(buff []byte, offset int) (int, error) {
 func (tun *NativeTun) GUID() windows.GUID {
 	return tun.wt.CfgInstanceID
 }
+
+//
+// GUID returns Windows adapter instance ID.
+//
+func (tun *NativeTun) LUID() uint64 {
+	return ((uint64(tun.wt.LUIDIndex) & ((1 << 24) - 1)) << 24) | ((uint64(tun.wt.IfType) & ((1 << 16) - 1)) << 48)
+}
