@@ -8,7 +8,6 @@ package wintun
 import (
 	"errors"
 	"fmt"
-	"golang.zx2c4.com/wireguard/tun/wintun/netshell"
 	"strings"
 	"syscall"
 	"time"
@@ -17,6 +16,8 @@ import (
 	"golang.org/x/sys/windows"
 	"golang.org/x/sys/windows/registry"
 	"golang.zx2c4.com/wireguard/tun/wintun/guid"
+	"golang.zx2c4.com/wireguard/tun/wintun/netshell"
+	registryEx "golang.zx2c4.com/wireguard/tun/wintun/registry"
 	"golang.zx2c4.com/wireguard/tun/wintun/setupapi"
 )
 
@@ -34,28 +35,23 @@ var deviceClassNetGUID = windows.GUID{Data1: 0x4d36e972, Data2: 0xe325, Data3: 0
 const hardwareID = "Wintun"
 const enumerator = ""
 const machineName = ""
+const waitForRegistryTimeout = time.Second * 5
 
 //
 // MakeWintun creates interface handle and populates it from device registry key
 //
-func MakeWintun(deviceInfoSet setupapi.DevInfo, deviceInfoData *setupapi.DevInfoData) (*Wintun, error) {
+func makeWintun(deviceInfoSet setupapi.DevInfo, deviceInfoData *setupapi.DevInfoData) (*Wintun, error) {
 	// Open HKEY_LOCAL_MACHINE\SYSTEM\CurrentControlSet\Control\Class\<class>\<id> registry key.
-	key, err := deviceInfoSet.OpenDevRegKey(deviceInfoData, setupapi.DICS_FLAG_GLOBAL, 0, setupapi.DIREG_DRV, registry.READ)
+	key, err := deviceInfoSet.OpenDevRegKey(deviceInfoData, setupapi.DICS_FLAG_GLOBAL, 0, setupapi.DIREG_DRV, registry.QUERY_VALUE)
 	if err != nil {
-		return nil, errors.New("Device-specific registry key open failed: " + err.Error())
+		return nil, fmt.Errorf("Device-specific registry key open failed: %v", err)
 	}
 	defer key.Close()
 
-	var valueStr string
-	var valueType uint32
-
 	// Read the NetCfgInstanceId value.
-	valueStr, valueType, err = keyGetStringValueRetry(key, "NetCfgInstanceId")
+	valueStr, err := registryEx.GetStringValue(key, "NetCfgInstanceId")
 	if err != nil {
-		return nil, errors.New("RegQueryStringValue(\"NetCfgInstanceId\") failed: " + err.Error())
-	}
-	if valueType != registry.SZ {
-		return nil, fmt.Errorf("NetCfgInstanceId registry value is not REG_SZ (expected: %v, provided: %v)", registry.SZ, valueType)
+		return nil, fmt.Errorf("RegQueryStringValue(\"NetCfgInstanceId\") failed: %v", err)
 	}
 
 	// Convert to windows.GUID.
@@ -65,15 +61,15 @@ func MakeWintun(deviceInfoSet setupapi.DevInfo, deviceInfoData *setupapi.DevInfo
 	}
 
 	// Read the NetLuidIndex value.
-	luidIdx, valueType, err := key.GetIntegerValue("NetLuidIndex")
+	luidIdx, _, err := key.GetIntegerValue("NetLuidIndex")
 	if err != nil {
-		return nil, errors.New("RegQueryValue(\"NetLuidIndex\") failed: " + err.Error())
+		return nil, fmt.Errorf("RegQueryValue(\"NetLuidIndex\") failed: %v", err)
 	}
 
 	// Read the NetLuidIndex value.
-	ifType, valueType, err := key.GetIntegerValue("*IfType")
+	ifType, _, err := key.GetIntegerValue("*IfType")
 	if err != nil {
-		return nil, errors.New("RegQueryValue(\"*IfType\") failed: " + err.Error())
+		return nil, fmt.Errorf("RegQueryValue(\"*IfType\") failed: %v", err)
 	}
 
 	return &Wintun{
@@ -99,7 +95,7 @@ func GetInterface(ifname string, hwndParent uintptr) (*Wintun, error) {
 	// Create a list of network devices.
 	devInfoList, err := setupapi.SetupDiGetClassDevsEx(&deviceClassNetGUID, enumerator, hwndParent, setupapi.DIGCF_PRESENT, setupapi.DevInfo(0), machineName)
 	if err != nil {
-		return nil, errors.New(fmt.Sprintf("SetupDiGetClassDevsEx(%v) failed: ", guid.ToString(&deviceClassNetGUID)) + err.Error())
+		return nil, fmt.Errorf("SetupDiGetClassDevsEx(%s) failed: %v", guid.ToString(&deviceClassNetGUID), err)
 	}
 	defer devInfoList.Close()
 
@@ -114,18 +110,19 @@ func GetInterface(ifname string, hwndParent uintptr) (*Wintun, error) {
 		// Get the device from the list. Should anything be wrong with this device, continue with next.
 		deviceData, err := devInfoList.EnumDeviceInfo(index)
 		if err != nil {
-			if errWin, ok := err.(syscall.Errno); ok && errWin == 259 /*ERROR_NO_MORE_ITEMS*/ {
+			if errWin, ok := err.(syscall.Errno); ok && errWin == windows.ERROR_NO_MORE_ITEMS {
 				break
 			}
 			continue
 		}
 
 		// Get interface ID.
-		wintun, err := MakeWintun(devInfoList, deviceData)
+		wintun, err := makeWintun(devInfoList, deviceData)
 		if err != nil {
 			continue
 		}
 
+		//TODO: is there a better way than comparing ifnames?
 		// Get interface name.
 		ifname2, err := wintun.GetInterfaceName()
 		if err != nil {
@@ -137,7 +134,7 @@ func GetInterface(ifname string, hwndParent uintptr) (*Wintun, error) {
 			const driverType = setupapi.SPDIT_COMPATDRIVER
 			err = devInfoList.BuildDriverInfoList(deviceData, driverType)
 			if err != nil {
-				return nil, errors.New("SetupDiBuildDriverInfoList failed: " + err.Error())
+				return nil, fmt.Errorf("SetupDiBuildDriverInfoList failed: %v", err)
 			}
 			defer devInfoList.DestroyDriverInfoList(deviceData, driverType)
 
@@ -145,7 +142,7 @@ func GetInterface(ifname string, hwndParent uintptr) (*Wintun, error) {
 				// Get a driver from the list.
 				driverData, err := devInfoList.EnumDriverInfo(deviceData, driverType, index)
 				if err != nil {
-					if errWin, ok := err.(syscall.Errno); ok && errWin == 259 /*ERROR_NO_MORE_ITEMS*/ {
+					if errWin, ok := err.(syscall.Errno); ok && errWin == windows.ERROR_NO_MORE_ITEMS {
 						break
 					}
 					// Something is wrong with this driver. Skip it.
@@ -191,52 +188,49 @@ func CreateInterface(description string, hwndParent uintptr) (*Wintun, bool, err
 	// Create an empty device info set for network adapter device class.
 	devInfoList, err := setupapi.SetupDiCreateDeviceInfoListEx(&deviceClassNetGUID, hwndParent, machineName)
 	if err != nil {
-		return nil, false, errors.New(fmt.Sprintf("SetupDiCreateDeviceInfoListEx(%v) failed: ", guid.ToString(&deviceClassNetGUID)) + err.Error())
+		return nil, false, fmt.Errorf("SetupDiCreateDeviceInfoListEx(%s) failed: %v", guid.ToString(&deviceClassNetGUID), err)
 	}
+	defer devInfoList.Close()
 
 	// Get the device class name from GUID.
 	className, err := setupapi.SetupDiClassNameFromGuidEx(&deviceClassNetGUID, machineName)
 	if err != nil {
-		return nil, false, errors.New(fmt.Sprintf("SetupDiClassNameFromGuidEx(%v) failed: ", guid.ToString(&deviceClassNetGUID)) + err.Error())
+		return nil, false, fmt.Errorf("SetupDiClassNameFromGuidEx(%s) failed: %v", guid.ToString(&deviceClassNetGUID), err)
 	}
 
 	// Create a new device info element and add it to the device info set.
 	deviceData, err := devInfoList.CreateDeviceInfo(className, &deviceClassNetGUID, description, hwndParent, setupapi.DICD_GENERATE_ID)
 	if err != nil {
-		return nil, false, errors.New("SetupDiCreateDeviceInfo failed: " + err.Error())
+		return nil, false, fmt.Errorf("SetupDiCreateDeviceInfo failed: %v", err)
 	}
 
 	// Set a device information element as the selected member of a device information set.
 	err = devInfoList.SetSelectedDevice(deviceData)
 	if err != nil {
-		return nil, false, errors.New("SetupDiSetSelectedDevice failed: " + err.Error())
+		return nil, false, fmt.Errorf("SetupDiSetSelectedDevice failed: %v", err)
 	}
 
 	// Set Plug&Play device hardware ID property.
-	hwid, err := syscall.UTF16FromString(hardwareID)
+	err = devInfoList.SetDeviceRegistryPropertyString(deviceData, setupapi.SPDRP_HARDWAREID, hardwareID)
 	if err != nil {
-		return nil, false, err // syscall.UTF16FromString(hardwareID) should never fail: hardwareID is const string without NUL chars.
-	}
-	err = devInfoList.SetDeviceRegistryProperty(deviceData, setupapi.SPDRP_HARDWAREID, setupapi.UTF16ToBuf(append(hwid, 0)))
-	if err != nil {
-		return nil, false, errors.New("SetupDiSetDeviceRegistryProperty(SPDRP_HARDWAREID) failed: " + err.Error())
+		return nil, false, fmt.Errorf("SetupDiSetDeviceRegistryProperty(SPDRP_HARDWAREID) failed: %v", err)
 	}
 
 	// Search for the driver.
-	const driverType = setupapi.SPDIT_CLASSDRIVER
-	err = devInfoList.BuildDriverInfoList(deviceData, driverType)
+	const driverType = setupapi.SPDIT_COMPATDRIVER
+	err = devInfoList.BuildDriverInfoList(deviceData, driverType) //TODO: This takes ~510ms
 	if err != nil {
-		return nil, false, errors.New("SetupDiBuildDriverInfoList failed: " + err.Error())
+		return nil, false, fmt.Errorf("SetupDiBuildDriverInfoList failed: %v", err)
 	}
 	defer devInfoList.DestroyDriverInfoList(deviceData, driverType)
 
 	driverDate := windows.Filetime{}
 	driverVersion := uint64(0)
-	for index := 0; ; index++ {
+	for index := 0; ; index++ { //TODO: This loop takes ~600ms
 		// Get a driver from the list.
 		driverData, err := devInfoList.EnumDriverInfo(deviceData, driverType, index)
 		if err != nil {
-			if errWin, ok := err.(syscall.Errno); ok && errWin == 259 /*ERROR_NO_MORE_ITEMS*/ {
+			if errWin, ok := err.(syscall.Errno); ok && errWin == windows.ERROR_NO_MORE_ITEMS {
 				break
 			}
 			// Something is wrong with this driver. Skip it.
@@ -273,7 +267,7 @@ func CreateInterface(description string, hwndParent uintptr) (*Wintun, bool, err
 	// Call appropriate class installer.
 	err = devInfoList.CallClassInstaller(setupapi.DIF_REGISTERDEVICE, deviceData)
 	if err != nil {
-		return nil, false, errors.New("SetupDiCallClassInstaller(DIF_REGISTERDEVICE) failed: " + err.Error())
+		return nil, false, fmt.Errorf("SetupDiCallClassInstaller(DIF_REGISTERDEVICE) failed: %v", err)
 	}
 
 	// Register device co-installers if any. (Ignore errors)
@@ -282,37 +276,104 @@ func CreateInterface(description string, hwndParent uintptr) (*Wintun, bool, err
 	// Install interfaces if any. (Ignore errors)
 	devInfoList.CallClassInstaller(setupapi.DIF_INSTALLINTERFACES, deviceData)
 
-	var wintun *Wintun
-	var rebootRequired bool
-
 	// Install the device.
 	err = devInfoList.CallClassInstaller(setupapi.DIF_INSTALLDEVICE, deviceData)
 	if err != nil {
-		err = errors.New("SetupDiCallClassInstaller(DIF_INSTALLDEVICE) failed: " + err.Error())
+		err = fmt.Errorf("SetupDiCallClassInstaller(DIF_INSTALLDEVICE) failed: %v", err)
 	}
 
+	var wintun *Wintun
+	var rebootRequired bool
+	var key registry.Key
+
 	if err == nil {
 		// Check if a system reboot is required. (Ignore errors)
 		if ret, _ := checkReboot(devInfoList, deviceData); ret {
 			rebootRequired = true
 		}
 
-		// Get network interface. DIF_INSTALLDEVICE returns almost immediately, while the device
-		// installation continues in the background. It might take a while, before all registry
+		// DIF_INSTALLDEVICE returns almost immediately, while the device installation
+		// continues in the background. It might take a while, before all registry
 		// keys and values are populated.
-		for numAttempts := 0; numAttempts < 30; numAttempts++ {
-			wintun, err = MakeWintun(devInfoList, deviceData)
-			if err != nil {
-				if errWin, ok := err.(syscall.Errno); ok && errWin == windows.ERROR_FILE_NOT_FOUND {
-					// Wait and retry. TODO: Wait for a cancellable event instead.
-					err = errors.New("Time-out waiting for adapter to get ready")
-					time.Sleep(time.Second / 4)
-					continue
-				}
+		const pollTimeout = time.Millisecond * 50
+		for i := 0; i < int(waitForRegistryTimeout/pollTimeout); i++ {
+			if i != 0 {
+				time.Sleep(pollTimeout)
+			}
+			key, err = devInfoList.OpenDevRegKey(deviceData, setupapi.DICS_FLAG_GLOBAL, 0, setupapi.DIREG_DRV, registry.QUERY_VALUE|registry.NOTIFY)
+			if err == nil {
+				break
 			}
-
-			break
 		}
+		if err == nil {
+			_, err = registryEx.GetStringValueWait(key, "NetCfgInstanceId", waitForRegistryTimeout)
+			if err == nil {
+				_, err = registryEx.GetIntegerValueWait(key, "NetLuidIndex", waitForRegistryTimeout)
+			}
+			if err == nil {
+				_, err = registryEx.GetIntegerValueWait(key, "*IfType", waitForRegistryTimeout)
+			}
+			key.Close()
+		}
+	}
+
+	if err == nil {
+		// Get network interface.
+		wintun, err = makeWintun(devInfoList, deviceData)
+	}
+
+	if err == nil {
+		// Wait for network registry key to emerge and populate.
+		key, err = registryEx.OpenKeyWait(
+			registry.LOCAL_MACHINE,
+			wintun.GetNetRegKeyName(),
+			registry.QUERY_VALUE|registry.NOTIFY,
+			waitForRegistryTimeout)
+		if err == nil {
+			_, err = registryEx.GetStringValueWait(key, "Name", waitForRegistryTimeout)
+			key.Close()
+		}
+	}
+
+	if err == nil {
+		// Wait for TCP/IP adapter registry key to emerge and populate.
+		key, err = registryEx.OpenKeyWait(
+			registry.LOCAL_MACHINE,
+			wintun.GetTcpipAdapterRegKeyName(), registry.QUERY_VALUE|registry.NOTIFY,
+			waitForRegistryTimeout)
+		if err == nil {
+			_, err = registryEx.GetStringValueWait(key, "IpConfig", waitForRegistryTimeout)
+			key.Close()
+		}
+	}
+
+	var tcpipInterfaceRegKeyName string
+	if err == nil {
+		tcpipInterfaceRegKeyName, err = wintun.GetTcpipInterfaceRegKeyName()
+		if err == nil {
+			// Wait for TCP/IP interface registry key to emerge.
+			key, err = registryEx.OpenKeyWait(
+				registry.LOCAL_MACHINE,
+				tcpipInterfaceRegKeyName, registry.QUERY_VALUE,
+				waitForRegistryTimeout)
+			if err == nil {
+				key.Close()
+			}
+		}
+	}
+
+	//
+	// All the registry keys and values we're relying on are present now.
+	//
+
+	if err == nil {
+		// Disable dead gateway detection on our interface.
+		key, err = registry.OpenKey(registry.LOCAL_MACHINE, tcpipInterfaceRegKeyName, registry.SET_VALUE)
+		if err != nil {
+			err = fmt.Errorf("Error opening interface-specific TCP/IP network registry key: %v", err)
+		}
+		key.SetDWordValue("EnableDeadGWDetect", 0)
+		key.Close()
 	}
 
 	if err == nil {
@@ -355,7 +416,7 @@ func (wintun *Wintun) DeleteInterface(hwndParent uintptr) (bool, bool, error) {
 	// Create a list of network devices.
 	devInfoList, err := setupapi.SetupDiGetClassDevsEx(&deviceClassNetGUID, enumerator, hwndParent, setupapi.DIGCF_PRESENT, setupapi.DevInfo(0), machineName)
 	if err != nil {
-		return false, false, errors.New(fmt.Sprintf("SetupDiGetClassDevsEx(%v) failed: ", guid.ToString(&deviceClassNetGUID)) + err.Error())
+		return false, false, fmt.Errorf("SetupDiGetClassDevsEx(%s) failed: %v", guid.ToString(&deviceClassNetGUID), err.Error())
 	}
 	defer devInfoList.Close()
 
@@ -364,14 +425,15 @@ func (wintun *Wintun) DeleteInterface(hwndParent uintptr) (bool, bool, error) {
 		// Get the device from the list. Should anything be wrong with this device, continue with next.
 		deviceData, err := devInfoList.EnumDeviceInfo(index)
 		if err != nil {
-			if errWin, ok := err.(syscall.Errno); ok && errWin == 259 /*ERROR_NO_MORE_ITEMS*/ {
+			if errWin, ok := err.(syscall.Errno); ok && errWin == windows.ERROR_NO_MORE_ITEMS {
 				break
 			}
 			continue
 		}
 
 		// Get interface ID.
-		wintun2, err := MakeWintun(devInfoList, deviceData)
+		//TODO: Store some ID in the Wintun object such that this call isn't required.
+		wintun2, err := makeWintun(devInfoList, deviceData)
 		if err != nil {
 			continue
 		}
@@ -386,13 +448,13 @@ func (wintun *Wintun) DeleteInterface(hwndParent uintptr) (bool, bool, error) {
 			// Set class installer parameters for DIF_REMOVE.
 			err = devInfoList.SetClassInstallParams(deviceData, &removeDeviceParams.ClassInstallHeader, uint32(unsafe.Sizeof(removeDeviceParams)))
 			if err != nil {
-				return false, false, errors.New("SetupDiSetClassInstallParams failed: " + err.Error())
+				return false, false, fmt.Errorf("SetupDiSetClassInstallParams failed: %v", err)
 			}
 
 			// Call appropriate class installer.
 			err = devInfoList.CallClassInstaller(setupapi.DIF_REMOVE, deviceData)
 			if err != nil {
-				return false, false, errors.New("SetupDiCallClassInstaller failed: " + err.Error())
+				return false, false, fmt.Errorf("SetupDiCallClassInstaller failed: %v", err)
 			}
 
 			// Check if a system reboot is required. (Ignore errors)
@@ -436,34 +498,33 @@ func checkReboot(deviceInfoSet setupapi.DevInfo, deviceInfoData *setupapi.DevInf
 // GetInterfaceName returns network interface name.
 //
 func (wintun *Wintun) GetInterfaceName() (string, error) {
-	key, err := registryOpenKeyRetry(registry.LOCAL_MACHINE, wintun.GetNetRegKeyName(), registry.QUERY_VALUE)
+	key, err := registry.OpenKey(registry.LOCAL_MACHINE, wintun.GetNetRegKeyName(), registry.QUERY_VALUE)
 	if err != nil {
-		return "", errors.New("Network-specific registry key open failed: " + err.Error())
+		return "", fmt.Errorf("Network-specific registry key open failed: %v", err)
 	}
 	defer key.Close()
 
 	// Get the interface name.
-	return getRegStringValue(key, "Name")
+	return registryEx.GetStringValue(key, "Name")
 }
 
 //
 // SetInterfaceName sets network interface name.
 //
 func (wintun *Wintun) SetInterfaceName(ifname string) error {
-	// We open the registry key before calling HrRename, because the registry open will wait until the key exists.
-	key, err := registryOpenKeyRetry(registry.LOCAL_MACHINE, wintun.GetNetRegKeyName(), registry.SET_VALUE)
-	if err != nil {
-		return errors.New("Network-specific registry key open failed: " + err.Error())
-	}
-	defer key.Close()
-
 	// We have to tell the various runtime COM services about the new name too. We ignore the
-	// error because netshell isn't available on servercore.
+	// error because netshell isn't available on servercore. It's also slow, so we run it in a
+	// separate thread.
 	// TODO: netsh.exe falls back to NciSetConnection in this case. If somebody complains, maybe
 	// we should do the same.
-	_ = netshell.HrRenameConnection(&wintun.CfgInstanceID, windows.StringToUTF16Ptr(ifname))
+	go netshell.HrRenameConnection(&wintun.CfgInstanceID, windows.StringToUTF16Ptr(ifname))
 
 	// Set the interface name. The above line should have done this too, but in case it failed, we force it.
+	key, err := registry.OpenKey(registry.LOCAL_MACHINE, wintun.GetNetRegKeyName(), registry.SET_VALUE)
+	if err != nil {
+		return fmt.Errorf("Network-specific registry key open failed: %v", err)
+	}
+	defer key.Close()
 	return key.SetStringValue("Name", ifname)
 }
 
@@ -471,35 +532,33 @@ func (wintun *Wintun) SetInterfaceName(ifname string) error {
 // GetNetRegKeyName returns interface-specific network registry key name.
 //
 func (wintun *Wintun) GetNetRegKeyName() string {
-	return fmt.Sprintf("SYSTEM\\CurrentControlSet\\Control\\Network\\%v\\%v\\Connection", guid.ToString(&deviceClassNetGUID), guid.ToString(&wintun.CfgInstanceID))
+	return fmt.Sprintf("SYSTEM\\CurrentControlSet\\Control\\Network\\%s\\%s\\Connection", guid.ToString(&deviceClassNetGUID), guid.ToString(&wintun.CfgInstanceID))
 }
 
 //
-// getRegStringValue function reads a string value from registry.
+// GetTcpipAdapterRegKeyName returns adapter-specific TCP/IP network registry key name.
 //
-// If the value type is REG_EXPAND_SZ the environment variables are expanded.
-// Should expanding fail, original string value and nil error are returned.
+func (wintun *Wintun) GetTcpipAdapterRegKeyName() string {
+	return fmt.Sprintf("SYSTEM\\CurrentControlSet\\Services\\Tcpip\\Parameters\\Adapters\\%s", guid.ToString(&wintun.CfgInstanceID))
+}
+
 //
-func getRegStringValue(key registry.Key, name string) (string, error) {
-	// Read string value.
-	value, valueType, err := keyGetStringValueRetry(key, name)
+// GetTcpipInterfaceRegKeyName returns interface-specific TCP/IP network registry key name.
+//
+func (wintun *Wintun) GetTcpipInterfaceRegKeyName() (path string, err error) {
+	key, err := registry.OpenKey(registry.LOCAL_MACHINE, wintun.GetTcpipAdapterRegKeyName(), registry.QUERY_VALUE)
 	if err != nil {
-		return "", err
+		return "", fmt.Errorf("Error opening adapter-specific TCP/IP network registry key: %v", err)
 	}
-
-	if valueType != registry.EXPAND_SZ {
-		// Value does not require expansion.
-		return value, nil
-	}
-
-	valueExp, err := registry.ExpandString(value)
+	paths, _, err := key.GetStringsValue("IpConfig")
+	key.Close()
 	if err != nil {
-		// Expanding failed: return original sting value.
-		return value, nil
+		return "", fmt.Errorf("Error reading IpConfig registry key: %v", err)
 	}
-
-	// Return expanded value.
-	return valueExp, nil
+	if len(paths) == 0 {
+		return "", errors.New("No TCP/IP interfaces found on adapter")
+	}
+	return fmt.Sprintf("SYSTEM\\CurrentControlSet\\Services\\%s", paths[0]), nil
 }
 
 //
